#!/usr/bin/perl -w
###############################################################################
# Purpose: automatically handles mailman administrative requests
# Usage:   run the script with --help option
# Author:  Vadim Zeitlin <vadim@wxwindows.org>
# License: BSD (http://opensource.org/licenses/bsd-license.php)
# Created: 28 September, 2002
# Version: 0.12
# RCS-Id:  $Id$
###############################################################################

use strict;

use Getopt::Long ();
use HTML::Entities ();
use HTML::HeadParser ();
use HTML::TokeParser ();
use HTTP::Request::Common qw(GET POST);
use HTTP::Cookies ();
use LWP::UserAgent ();
use LWP::Simple ();

# uncomment this for tons of debugging messages
#use LWP::Debug qw(+);


###############################################################################
# Subroutines
###############################################################################

sub ProcessMessages();
sub ListMessages();
sub ReadConfig($);


# return the URL we're redirected to
sub GetRedirectUrl($) {
    my $response = $_[0];

    my $url = $response->header('Location') || return undef;

    # the Location URL is sometimes non-absolute which is not allowed, fix it
    local $URI::ABS_ALLOW_RELATIVE_SCHEME = 1;
    my $base = $response->base;
    $url = $HTTP::URI_CLASS->new($url, $base)->abs($base);

    return $url;
}

# go to the given tag, die if not found
#
# parameters: HTML::TokeParser object, the tag to go to, HTML text
# returns: the array reference returned by (successful) get_tag()
sub GoToTag($$$$) {
    my ($htmlParser, $tag, $html, $what) = @_;

    my $token = $htmlParser->get_tag($tag);
    if ( !$token ) {
        open(BAD, ">bad.html") or die;
        print BAD $html;
        die "Tag <$tag> not found when looking for $what, page dumped to bad.html for examination.\n";
    }

    return $token;
}

# extract the list of the messages being held for approval from the mailman
# administrative tasks HTML page
#
# parameters: the full HTML text of the page
# returns: array of (msgno, from, subject, reason) arrays
sub GetListOfHeldMessages($) {
    my $html = $_[0];

    # look whether we have any messages to approve
    my $marker = "Posting Held for Approval";
    my $pos = index $html, $marker;
    if ( $pos == -1 ) {
        # no messages at all
        return;
    }

    # and how many of them do we have?
    my $count;
    if ( substr($html, $pos + length $marker, 30) =~ /^ \(1 of (\d+)\)/ ) {
        $count = $1;
    }
    else {
        # mailman omits "1 of N" if there is only one message
        $count = 1;
    }

    # now extract the information about all messages
    my @msgs = ();
    my $htmlParser = HTML::TokeParser->new(\$html);

    # locate the beginning of the form
    GoToTag($htmlParser, "form", $html, "start of the form");

    # extract the message details: they're given in the first 3 table rows
    for ( 1..$count ) {
        GoToTag($htmlParser, "table", $html, "start of the table");
        GoToTag($htmlParser, "td", $html, "from row");
        GoToTag($htmlParser, "td", $html, "from value");
        my $from = $htmlParser->get_text("/td");

        GoToTag($htmlParser, "td", $html, "subject row");
        GoToTag($htmlParser, "td", $html, "subject value");
        my $subject = $htmlParser->get_text("/td");

        GoToTag($htmlParser, "td", $html, "reason row");
        GoToTag($htmlParser, "td", $html, "reason value");
        my $reason = $htmlParser->get_text("/td");

        # the number of the message
        my $number = GoToTag($htmlParser, "input", $html, "radiobutton")->[1]{'name'};

        GoToTag($htmlParser, "textarea", $html, "reject message");
        GoToTag($htmlParser, "textarea", $html, "headers text");
        my $headers = $htmlParser->get_text("/textarea");
        GoToTag($htmlParser, "textarea", $html, "contents text");
        my $content = $htmlParser->get_text("/textarea");
        push @msgs, [$number, $from, $subject, $reason, $headers, $content];
    }

    return @msgs;
}

###############################################################################
# Main program
###############################################################################

# command line options parsing
###############################################################################

my $VERSION_STRING = "0.12";

# boolean: if true, suppress messages
my $quiet = 0;

# the base admindb URL
my $baseurl;

# the list of messages to operate on
my @messages;

# the command to perform
my $command;

# the indices of the actions corresponding to different commands
my %commands = (
        "" => -1,
        "list" => -1,
        "accept" => 1,
        "reject" => 2,
        "discard" => 3,
        "filter" => 3       # same as discard
    );

# the admin password
my $password;

# file to store cookies in
my $cookie_file = '';

# the proxy to use
my $proxy = "";

# booleans: version/help requested from cmd line?
my $version = 0;
my $help = 0;

# boolean: only simulate actions?
my $dryrun = 0;

my $usage = <<EOF
$0 [options] <admindb_url> [<command> [message[, message...]]]

where command is one of: list (default), accept, discard, reject or filter and
is applied to the messages with the specified numbers (or all by default) held
for approval at the given mailman URL.

--help          give the usage message showing the program options
--version       show the program version and exit
--verbose       give verbose informational messages (default)
--quiet         be silent, only error messages are given
--dry-run       don't really discard/reject messages, just simulate
--pass=pass     the mailing list administrative password
--cookies=xxx   file to use to store cookies in
--proxy=url     use the given proxy, if 'no' don't use proxy at all (even not
                the environment variable http_proxy which is used by default)
EOF
;

# variables read from the config file
###############################################################################

# the exclude patterns for "From:"
my @from = ();

# for "Subject:"
my @subject = ();

# for mailman reason string
my @reason = ();

# for the message headers
my @header = ();

# for the (excerpt of) message content
my @content = ();


# main program start
###############################################################################

# don't forget to modify the usage section above if you add more options!
Getopt::Long::GetOptions
(
    'verbose'       => sub { $quiet = 0 },
    'quiet'         => \$quiet,
    'version'       => \$version,
    'help'          => \$help,
    'dry-run|n'     => \$dryrun,
    'pass=s'        => \$password,
    'cookies=s'     => \$cookie_file,
    'proxy=s'       => \$proxy,
    '<>'            => sub {
                        my ($value) = @_;
                        # The value is an object, not a string, stringify it
                        # explicitly to avoid problems when passing baseurl
                        # reference to GET.
                        $value .= "";
                        if ( $baseurl ) {
                            if ( $command ) {
                                push @messages, split(/,/, $value)
                            }
                            else {
                                $command = lc $value
                            }
                        }
                        else {
                            $baseurl = $value
                        }
                       }
) || die $usage;

if ( $version || $help ) {
    if ( $version ) {
        print "mladmin version $VERSION_STRING, " .
              "written by Vadim Zeitlin <vadim\@wxwindows.org>\n";
    }
    else {
        print $usage;
    }

    exit 0;
}

if ( !$baseurl ) {
    die "Required argument \"admindb_url\" is missing.\n$usage"
}

$command = "" if !defined $command;
if ( !defined($commands{$command}) ) {
    die "Unknown command \"$command\": only \"list\", \"accept\", " .
        "\"discard\" or \"reject\" allowed.\n" . "$usage"
}

# although the config file could be also used for other things, currently it
# is only useful when filtering so save efforts when we don't do this
if ( $command eq "filter" ) {
    my $conffile = "mladmin";

    if ( $^O !~ 'Win32' ) {
        # global config file only makes sense under Unix
        ReadConfig "/etc/mladmin.conf";

        # and the user config file is in HOME in this case
        $conffile = $ENV{"HOME"} . "/." . $conffile . ".conf";
    }
    else { # Windows
        $conffile .= ".ini";
    }

    ReadConfig $conffile
}

# the short list name
my $list = $baseurl;
$list =~ s@^.*admindb/@@;

# we need to "login" to the service first: for this we need to get the cookie
# which is set by the page we are redirected to when we initially try to
# access the archive
my $ua = LWP::UserAgent->new;
$ua->agent("mladmin/$VERSION_STRING");
if ( $proxy ) {
    $ua->proxy('http', $proxy);
}
elsif ( $proxy ne 'no' ) {
    # use http_proxy env var if set
    $ua->env_proxy();
}
#else: don't use proxy at all

# determine what to do with cookies
my $cookie_jar;
if ( $cookie_file ) {
    # store them
    $cookie_jar = HTTP::Cookies->new(file => $cookie_file,
                                     autosave => 1,
                                     ignore_discard => 1);
}
else {
    # don't store, use just during this session
    $cookie_jar = HTTP::Cookies->new();
}
$ua->cookie_jar($cookie_jar);

# unbuffer stdout as we want the progress messages to appear immediately
select((select(STDOUT), $| = 1)[0]);

# try to get the list of admin tasks directly if we use an existing cookies
# file (otherwise wecan be sure that we need to login first)
my $html;   # the text of the admindb main page
my $login;  # do we need to login?
if ( $cookie_file ) {
    my $request = GET $baseurl;
    my $response = $ua->simple_request($request);
    if ( !$response->is_success ) {
        die "Failed to retrieve the list of administrative tasks.\n"
    }

    $html = $response->content;

    # check if we really logged in, the cookie might be expired or invalid in
    # which case we'd be at the login page instead (with title "Administrative
    # Authentication" for older mailman versions or "Moderator Authentication"
    # for mailman 2.1+
    $login = $html =~ /(Administrative|Moderator) Authentication/;
}
else {
    $login = 1;
}

if ( $login ) {
    if ( !$password ) {
        my $hasReadKey = eval { require Term::ReadKey; };
        if ( !$hasReadKey ) {
            warn "Please install Term::ReadKey to avoid echoing the password!\n";
        }

        print "Administrative password for $list: ";

        if ( $hasReadKey ) {
            Term::ReadKey::ReadMode('noecho');
            $password = Term::ReadKey::ReadLine(0); # 0 means normal read
            Term::ReadKey::ReadMode('restore');
        }
        else {
            $password = <STDIN>
        }

        chomp $password;

        print "\n";
    }

    print "Logging in to $list... " unless $quiet;

    my $request = POST $baseurl, [ "adminpw" => $password ];
    my $response = $ua->simple_request($request);

    # although simple_request() adds the cookies for us we should
    # extract them from the response we get manually
    $cookie_jar->extract_cookies($response);

    while ( $response->is_redirect ) {
        my $url = GetRedirectUrl($response);

        # go to the new page
        if ( $url ) {
            $request = GET $url;
            $response = $ua->simple_request($request);
        }
    }

    if ( !$response->is_success ) {
        print "Failed!\n";
        exit 2
    }

    $html = $response->content;

    if ( $html =~ /Authorization\s+failed/s ) {
        print "wrong password\n";
        exit 3
    }

    print "ok.\n";
}
#else: already authentificated

# check if we have mailman >= 2.1 which doesn't provide all the message details
# at $baseurl but has a link to the details page instead, before the main form
# start
my $htmlParser = HTML::TokeParser->new(\$html);
my $detailsurl = "$baseurl?details=all";
while ( my $token = $htmlParser->get_tag('a', 'form') ) {
    # no link, continue with this page
    last if ( $token->[0] eq 'form' );

    if ( $token->[1]{'href'} eq $detailsurl ) {
        # this is the link to the page we need, follow it
        my $response = $ua->simple_request(GET $detailsurl);
        if ( !$response->is_success ) {
            print "Failed to retrieve the details page $detailsurl.\n";
            exit 2
        }

        $html = $response->content;
    }
}

# get the list of all postings held for approval
my @msgs = GetListOfHeldMessages($html);
my $count = $#msgs + 1;
if ( !$count ) {
    warn "No held messages for the list $list.\n";

    exit 1
}

# process the command given to us
if ( $command eq "" || $command eq "list" ) {
    ListMessages
}
else { # command == "discard" or "accept" or "reject"
    ProcessMessages
}

exit 0;

###############################################################################
# More subroutines
###############################################################################

# parse the config file with given name
sub ReadConfig($) {
    my $conffile = $_[0];

    open(CONFIG, $conffile) or return 0;
    while (<CONFIG>) {
        # skip empty line and comments
        chomp;
        next if /^#/;
        next if /^\s*$/;

        # analyze the line of the form "<cmd> <key> <value>"
        my ($cmd, $key, $value) = split ' ', $_, 3;
        if ( $cmd eq 'reject' ) {
            if ( $key =~ 'from' ) {
                push @from, $value;
            }
            elsif ( $key =~ 'subject' ) {
                push @subject, $value;
            }
            elsif ( $key =~ 'header' ) {
                push @header, $value;
            }
            elsif ( $key =~ 'content' ) {
                push @content, $value;
            }
            elsif ( $key =~ 'reason' ) {
                push @reason, $value;
            }
            else {
                warn "$conffile($.): unknown key \"$key\" ignored.\n";
            }
        }
        else {
            warn "$conffile($.): unknown command \"$cmd\" ignored.\n";
        }
    }

    return 1;
}


# check if the message should be filtered
sub ShouldFilterMessage($) {
    # check if the message matches a single criterium
    sub CheckCriterium($$$$) {
        my ($msg, $needle, $haystack, $field) = @_;
        foreach ( @{$haystack} ) {
            if ( $needle =~ $_ ) {
                print "\n\tMessage " . @{$msg}[0] . ": " .
                      "$field matches \"$_\", discarded." unless $quiet;

                return 1
            }
        }

        return 0
    }

    my @msg = @{$_[0]};

    return CheckCriterium(\@msg, $msg[1], \@from, "sender") ||
           CheckCriterium(\@msg, $msg[2], \@subject, "subject") ||
           CheckCriterium(\@msg, $msg[3], \@reason, "mailman reason") ||
           CheckCriterium(\@msg, $msg[4], \@header, "header") ||
           CheckCriterium(\@msg, $msg[1], \@content, "contents")
}

# discard, reject, accept or filter the messages
sub ProcessMessages() {
    my $all = $#messages == -1;
    print "Preparing to $command " . ($all ? "all " : "") . "message" .
          (!$all && $#messages == 0 ? "" : "s") . " " .
          ($all ? "" : join(',', @messages)) . "... " unless $quiet;

    # build the POST arguments string
    my %args;

    for ( $all ? 0..$#msgs : map(--$_, @messages) ) {
        if ( $command eq "filter" ) {
            # check if we should filter this message
            next if !ShouldFilterMessage($msgs[$_])
        }

        # the radiobox command index
        $args{@{$msgs[$_]}[0]} = $commands{$command}
    }

    if ( !scalar(%args) ) {
        print "nothing to do.\n" unless $quiet;
    }
    else {
        my $count;
        if ( $command eq "filter" ) {
            $count = scalar(keys %args);
            print "\nDiscarding messages... " unless $quiet;

            # adjust for the message below
            $command = "discard"
        }
        else {
            $count = ($all ? $#msgs : $#messages) + 1;
        }

        if ( !$dryrun ) {
            my $request = POST $baseurl, \%args;
            my $response = $ua->simple_request($request);
            if ( !$response->is_success ) {
                warn "failed!\n";
                return
            }
        }

        my $sfx = $count == 1 ? "" : "s";
        printf "$count message$sfx " . ($dryrun ? "would be " : "") .
                    "${command}ed.\n" unless $quiet;
    }
}

# list the messages and optionally propose to discard them
sub ListMessages() {
    # get the terminal size: use Term::Size if installed, fall back to
    # environment otherwise
    my $COLUMNS = 0;
    if ( eval { require Term::Size; } ) {
        $COLUMNS = &Term::Size::chars
    }

    $COLUMNS ||= $ENV{"COLUMNS"} || 80;

    print "$count message" . ($count == 1 ? "" : "s") .
          " held for approval for the list $list:\n" .
          "\n";

    my ($widthFrom, $widthSubj, $widthReason) =
        ($COLUMNS / 3, $COLUMNS / 3, $COLUMNS - 2*($COLUMNS / 3) - 5);

    my $fmt = sprintf "%%-%d.%ds%%-%d.%ds%%-%d.%ds\n",
                      $widthFrom + 1, $widthFrom,
                      $widthSubj + 1, $widthSubj,
                      $widthReason, $widthReason;

    printf " N " . $fmt,
           "From", "Subject", "Reason";

    print '-' x $COLUMNS . "\n";

    for (0..$#msgs) {
        my @msg = @{$msgs[$_]};
        printf "%2d " . $fmt,
               $_ + 1, $msg[1], $msg[2], $msg[3]
    }

    print '-' x $COLUMNS . "\n";

    # don't propose to discard messages if "--list" was explicitly specified
    if ( $command eq "" ) {
        print "Discard all messages (y/[n])? ";
        if ( <> =~ /^y/i ) {
            $command = "discard";
            ProcessMessages
        }
        else {
            print "Nothing done.\n" unless $quiet
        }
    }
}

# History:
#
# 0.04:
#       added reject support (Ethan Galstad)
#       added filter command (Daniel Veillard)
#
# 0.03:
#       ask whether to delete all messages
#       better support for terminals with width != 80
#
# 0.02:
#       added "messages" argument and accept command
#
# 0.01:
#       initial barely functional release

# vi: set ts=4 sw=4 et cin:
